@@ -55,6 +55,7 @@ enum connman_network_error {
 	CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL  = 2,
 	CONNMAN_NETWORK_ERROR_INVALID_KEY     = 3,
 	CONNMAN_NETWORK_ERROR_CONNECT_FAIL    = 4,
+	CONNMAN_NETWORK_ERROR_BLOCKED         = 5,
 };
 
 #define CONNMAN_NETWORK_PRIORITY_LOW      -100
@@ -102,6 +103,11 @@ bool connman_network_get_connected(struct connman_network *network);
 
 bool connman_network_get_associating(struct connman_network *network);
 
+bool connman_network_get_connectable(struct connman_network *network);
+
+int connman_network_set_connectable(struct connman_network *network,
+		bool connectable);
+
 void connman_network_clear_hidden(void *user_data);
 int connman_network_connect_hidden(struct connman_network *network,
 			char *identity, char* passphrase, void *user_data);
