@@ -50,6 +50,7 @@ struct connman_network {
 	bool available;
 	bool connected;
 	bool roaming;
+	bool connectable;
 	uint8_t strength;
 	uint16_t frequency;
 	char *identifier;
@@ -81,6 +82,10 @@ struct connman_network {
 		char *anonymous_identity;
 		char *agent_identity;
 		char *ca_cert_path;
+		char *subject_match;
+		char *altsubject_match;
+		char *domain_suffix_match;
+		char *domain_match;
 		char *client_cert_path;
 		char *private_key_path;
 		char *private_key_passphrase;
@@ -136,8 +141,6 @@ static void set_configuration(struct connman_network *network,
 
 	__connman_device_set_network(network->device, network);
 
-	connman_device_set_disconnected(network->device, false);
-
 	service = connman_service_lookup_from_network(network);
 	__connman_service_ipconfig_indicate_state(service,
 					CONNMAN_SERVICE_STATE_CONFIGURATION,
@@ -287,13 +290,8 @@ static int manual_ipv6_set(struct connman_network *network,
 	if (err < 0)
 		return err;
 
-	__connman_connection_gateway_activate(service,
-						CONNMAN_IPCONFIG_TYPE_IPV6);
-
 	__connman_device_set_network(network->device, network);
 
-	connman_device_set_disconnected(network->device, false);
-
 	connman_network_set_associating(network, false);
 
 	network->connecting = false;
@@ -558,8 +556,6 @@ static void autoconf_ipv6_set(struct connman_network *network)
 
 	__connman_device_set_network(network->device, network);
 
-	connman_device_set_disconnected(network->device, false);
-
 	service = connman_service_lookup_from_network(network);
 	if (!service)
 		return;
@@ -804,23 +800,6 @@ static void network_remove(struct connman_network *network)
 	network->driver = NULL;
 }
 
-static void network_change(struct connman_network *network)
-{
-	DBG("network %p name %s", network, network->name);
-
-	if (!network->connected)
-		return;
-
-	connman_device_set_disconnected(network->device, true);
-
-	if (network->driver && network->driver->disconnect) {
-		network->driver->disconnect(network);
-		return;
-	}
-
-	network->connected = false;
-}
-
 static void probe_driver(struct connman_network_driver *driver)
 {
 	GSList *list;
@@ -843,20 +822,6 @@ static void probe_driver(struct connman_network_driver *driver)
 	}
 }
 
-static void remove_driver(struct connman_network_driver *driver)
-{
-	GSList *list;
-
-	DBG("driver %p name %s", driver, driver->name);
-
-	for (list = network_list; list; list = list->next) {
-		struct connman_network *network = list->data;
-
-		if (network->driver == driver)
-			network_remove(network);
-	}
-}
-
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	const struct connman_network_driver *driver1 = a;
@@ -865,6 +830,18 @@ static gint compare_priority(gconstpointer a, gconstpointer b)
 	return driver2->priority - driver1->priority;
 }
 
+int connman_network_set_connectable(struct connman_network *network,
+		bool connectable)
+{
+	network->connectable = connectable;
+	return 0;
+}
+
+bool connman_network_get_connectable(struct connman_network *network)
+{
+	return network->connectable;
+}
+
 /**
  * connman_network_driver_register:
  * @driver: network driver definition
@@ -893,11 +870,18 @@ int connman_network_driver_register(struct connman_network_driver *driver)
  */
 void connman_network_driver_unregister(struct connman_network_driver *driver)
 {
+	GSList *list;
+
 	DBG("driver %p name %s", driver, driver->name);
 
 	driver_list = g_slist_remove(driver_list, driver);
 
-	remove_driver(driver);
+	for (list = network_list; list; list = list->next) {
+		struct connman_network *network = list->data;
+
+		if (network->driver == driver)
+			network_remove(network);
+	}
 }
 
 static void network_destruct(struct connman_network *network)
@@ -913,6 +897,10 @@ static void network_destruct(struct connman_network *network)
 	g_free(network->wifi.anonymous_identity);
 	g_free(network->wifi.agent_identity);
 	g_free(network->wifi.ca_cert_path);
+	g_free(network->wifi.subject_match);
+	g_free(network->wifi.altsubject_match);
+	g_free(network->wifi.domain_suffix_match);
+	g_free(network->wifi.domain_match);
 	g_free(network->wifi.client_cert_path);
 	g_free(network->wifi.private_key_path);
 	g_free(network->wifi.private_key_passphrase);
@@ -944,14 +932,10 @@ struct connman_network *connman_network_create(const char *identifier,
 	struct connman_network *network;
 	char *ident;
 
-	DBG("identifier %s type %d", identifier, type);
-
 	network = g_try_new0(struct connman_network, 1);
 	if (!network)
 		return NULL;
 
-	DBG("network %p", network);
-
 	network->refcount = 1;
 
 	ident = g_strdup(identifier);
@@ -966,6 +950,8 @@ struct connman_network *connman_network_create(const char *identifier,
 
 	network_list = g_slist_prepend(network_list, network);
 
+	DBG("network %p identifier %s type %s", network, identifier,
+		type2string(type));
 	return network;
 }
 
@@ -1270,6 +1256,16 @@ static void set_connect_error(struct connman_network *network)
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
 }
 
+static void set_blocked_error(struct connman_network *network)
+{
+	struct connman_service *service;
+
+	service = connman_service_lookup_from_network(network);
+
+	__connman_service_indicate_error(service,
+					CONNMAN_SERVICE_ERROR_BLOCKED);
+}
+
 void connman_network_set_ipv4_method(struct connman_network *network,
 					enum connman_ipconfig_method method)
 {
@@ -1309,9 +1305,6 @@ void connman_network_set_error(struct connman_network *network,
 {
 	DBG("network %p error %d", network, error);
 
-	network->connecting = false;
-	network->associating = false;
-
 	switch (error) {
 	case CONNMAN_NETWORK_ERROR_UNKNOWN:
 		return;
@@ -1327,9 +1320,12 @@ void connman_network_set_error(struct connman_network *network,
 	case CONNMAN_NETWORK_ERROR_CONNECT_FAIL:
 		set_connect_error(network);
 		break;
+	case CONNMAN_NETWORK_ERROR_BLOCKED:
+		set_blocked_error(network);
+		break;
 	}
 
-	network_change(network);
+	__connman_network_disconnect(network);
 }
 
 /**
@@ -1350,8 +1346,7 @@ int connman_network_set_connected(struct connman_network *network,
 							!connected) {
 		connman_network_set_error(network,
 					CONNMAN_NETWORK_ERROR_CONNECT_FAIL);
-		if (__connman_network_disconnect(network) == 0)
-			return 0;
+		return 0;
 	}
 
 	if (network->connected == connected)
@@ -1742,8 +1737,6 @@ int connman_network_set_name(struct connman_network *network,
 int connman_network_set_strength(struct connman_network *network,
 						uint8_t strength)
 {
-	DBG("network %p strengh %d", network, strength);
-
 	network->strength = strength;
 
 	return 0;
@@ -1757,8 +1750,6 @@ uint8_t connman_network_get_strength(struct connman_network *network)
 int connman_network_set_frequency(struct connman_network *network,
 						uint16_t frequency)
 {
-	DBG("network %p frequency %d", network, frequency);
-
 	network->frequency = frequency;
 
 	return 0;
@@ -1772,8 +1763,6 @@ uint16_t connman_network_get_frequency(struct connman_network *network)
 int connman_network_set_wifi_channel(struct connman_network *network,
 						uint16_t channel)
 {
-	DBG("network %p wifi channel %d", network, channel);
-
 	network->wifi.channel = channel;
 
 	return 0;
@@ -1795,8 +1784,6 @@ uint16_t connman_network_get_wifi_channel(struct connman_network *network)
 int connman_network_set_string(struct connman_network *network,
 					const char *key, const char *value)
 {
-	DBG("network %p key %s value %s", network, key, value);
-
 	if (g_strcmp0(key, "Name") == 0)
 		return connman_network_set_name(network, value);
 
@@ -1830,6 +1817,18 @@ int connman_network_set_string(struct connman_network *network,
 	} else if (g_str_equal(key, "WiFi.CACertFile")) {
 		g_free(network->wifi.ca_cert_path);
 		network->wifi.ca_cert_path = g_strdup(value);
+	} else if (g_str_equal(key, "WiFi.SubjectMatch")) {
+		g_free(network->wifi.subject_match);
+		network->wifi.subject_match = g_strdup(value);
+	} else if (g_str_equal(key, "WiFi.AltSubjectMatch")) {
+		g_free(network->wifi.altsubject_match);
+		network->wifi.altsubject_match = g_strdup(value);
+	} else if (g_str_equal(key, "WiFi.DomainSuffixMatch")) {
+		g_free(network->wifi.domain_suffix_match);
+		network->wifi.domain_suffix_match = g_strdup(value);
+	} else if (g_str_equal(key, "WiFi.DomainMatch")) {
+		g_free(network->wifi.domain_match);
+		network->wifi.domain_match = g_strdup(value);
 	} else if (g_str_equal(key, "WiFi.ClientCertFile")) {
 		g_free(network->wifi.client_cert_path);
 		network->wifi.client_cert_path = g_strdup(value);
@@ -1862,8 +1861,6 @@ int connman_network_set_string(struct connman_network *network,
 const char *connman_network_get_string(struct connman_network *network,
 							const char *key)
 {
-	DBG("network %p key %s", network, key);
-
 	if (g_str_equal(key, "Path"))
 		return network->path;
 	else if (g_str_equal(key, "Name"))
@@ -1886,6 +1883,14 @@ const char *connman_network_get_string(struct connman_network *network,
 		return network->wifi.agent_identity;
 	else if (g_str_equal(key, "WiFi.CACertFile"))
 		return network->wifi.ca_cert_path;
+	else if (g_str_equal(key, "WiFi.SubjectMatch"))
+		return network->wifi.subject_match;
+	else if (g_str_equal(key, "WiFi.AltSubjectMatch"))
+		return network->wifi.altsubject_match;
+	else if (g_str_equal(key, "WiFi.DomainSuffixMatch"))
+		return network->wifi.domain_suffix_match;
+	else if (g_str_equal(key, "WiFi.DomainMatch"))
+		return network->wifi.domain_match;
 	else if (g_str_equal(key, "WiFi.ClientCertFile"))
 		return network->wifi.client_cert_path;
 	else if (g_str_equal(key, "WiFi.PrivateKeyFile"))
@@ -1911,8 +1916,6 @@ const char *connman_network_get_string(struct connman_network *network,
 int connman_network_set_bool(struct connman_network *network,
 					const char *key, bool value)
 {
-	DBG("network %p key %s value %d", network, key, value);
-
 	if (g_strcmp0(key, "Roaming") == 0)
 		network->roaming = value;
 	else if (g_strcmp0(key, "WiFi.WPS") == 0)
@@ -1933,8 +1936,6 @@ int connman_network_set_bool(struct connman_network *network,
 bool connman_network_get_bool(struct connman_network *network,
 							const char *key)
 {
-	DBG("network %p key %s", network, key);
-
 	if (g_str_equal(key, "Roaming"))
 		return network->roaming;
 	else if (g_str_equal(key, "WiFi.WPS"))
@@ -1957,8 +1958,6 @@ bool connman_network_get_bool(struct connman_network *network,
 int connman_network_set_blob(struct connman_network *network,
 			const char *key, const void *data, unsigned int size)
 {
-	DBG("network %p key %s size %d", network, key, size);
-
 	if (g_str_equal(key, "WiFi.SSID")) {
 		g_free(network->wifi.ssid);
 		network->wifi.ssid = g_try_malloc(size);
@@ -1985,8 +1984,6 @@ int connman_network_set_blob(struct connman_network *network,
 const void *connman_network_get_blob(struct connman_network *network,
 					const char *key, unsigned int *size)
 {
-	DBG("network %p key %s", network, key);
-
 	if (g_str_equal(key, "WiFi.SSID")) {
 		if (size)
 			*size = network->wifi.ssid_len;
